import numpy as np
def _get_pairwise_all_data(dataset):
    user_input, item_input_pos,item_input_neg = [], [], []
    trainMatrix = dataset.trainMatrix
    num_items = dataset.num_items
    for (u, i) in trainMatrix.keys():
        user_input.append(u)
        item_input_pos.append(i)
        j = np.random.randint(num_items)
        while (u, j) in trainMatrix.keys():
            j = np.random.randint(num_items)
        item_input_neg.append(j)
    user_input = np.array(user_input, dtype=np.int32)
    item_input_pos = np.array(item_input_pos, dtype=np.int32)
    item_input_neg = np.array(item_input_neg, dtype=np.int32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input=user_input[shuffle_index]
    item_input_pos=item_input_pos[shuffle_index]
    item_input_neg=item_input_neg[shuffle_index]
    return user_input, item_input_pos,item_input_neg 

def _get_pairwise_all_highorder_data(dataset,high_order):
    user_input, item_input_pos,item_input_recents,item_input_neg = [], [], [],[]
    trainMatrix = dataset.trainMatrix
    num_items = dataset.num_items
    num_users = dataset.num_users
    trainDict = dataset.trainDict
    for u in range(num_users):
        items_by_user = trainDict[u]
        for idx in range(high_order,len(items_by_user)):
            i = items_by_user[idx] # item id 
            # positive instance
            user_input.append(u)
            item_input_pos.append(i)
            item_input_recent = []
            for t in range(1,high_order+1):
                item_input_recent.append(items_by_user[idx-t])
            item_input_recents.append(item_input_recent)
            j = np.random.randint(num_items)
            while (u, j) in trainMatrix.keys():
                j = np.random.randint(num_items)
            item_input_neg.append(j)
    user_input = np.array(user_input, dtype=np.int32)
    item_input_pos = np.array(item_input_pos, dtype=np.int32)
    item_input_recents = np.array(item_input_recents)
    item_input_neg = np.array(item_input_neg, dtype=np.int32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input = user_input[shuffle_index]
    item_input_pos = item_input_pos[shuffle_index]
    item_input_recents = item_input_recents[shuffle_index]
    item_input_neg = item_input_neg[shuffle_index]    
    return user_input, item_input_pos,item_input_recents,item_input_neg 

def _get_pairwise_all_firstorder_data(dataset):
    user_input, item_input_pos,item_input_recent,item_input_neg = [], [], [],[]
    trainMatrix = dataset.trainMatrix
    num_items = dataset.num_items
    num_users = dataset.num_users
    train_dict = dataset.trainDict
    for u in range(num_users):
        items_by_user = train_dict[u]
        for idx in range(1,len(items_by_user)):
            i = items_by_user[idx] # item id 
            # positive instance
            user_input.append(u)
            item_input_pos.append(i)
            item_input_recent.append(items_by_user[idx-1])
            j = np.random.randint(num_items)
            while (u, j) in trainMatrix.keys():
                j = np.random.randint(num_items)
            item_input_neg.append(j)
    user_input = np.array(user_input, dtype=np.int32)
    item_input_pos = np.array(item_input_pos, dtype=np.int32)
    item_input_recent = np.array(item_input_recent,dtype=np.int32)
    item_input_neg = np.array(item_input_neg, dtype=np.int32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input = user_input[shuffle_index]
    item_input_pos = item_input_pos[shuffle_index]
    item_input_recent = item_input_recent[shuffle_index]
    item_input_neg = item_input_neg[shuffle_index]    
    return user_input,item_input_pos,item_input_recent,item_input_neg 

def _get_pairwise_all_likefism_data(dataset):
    user_input_pos,user_input_neg, num_idx_pos, num_idx_neg, item_input_pos,item_input_neg = [], [], [],[],[],[]
    trainMatrix = dataset.trainMatrix
    num_items = dataset.num_items
    num_users = dataset.num_users
    trainDict = dataset.trainDict
    for u in range(num_users):
        items_by_user = trainDict[u].copy()
        size = len(items_by_user)   
        for i in items_by_user:
            j = np.random.randint(num_items)
            while (u,j) in trainMatrix.keys():
                j = np.random.randint(num_items)
            user_input_neg.append(items_by_user)
            num_idx_neg.append(size)
            item_input_neg.append(j)
            
            items_by_user.remove(i)
            user_input_pos.append(items_by_user)
            num_idx_pos.append(size-1)
            item_input_pos.append(i)
    user_input_pos = np.array(user_input_pos)
    user_input_neg = np.array(user_input_neg)
    num_idx_pos = np.array(num_idx_pos,dtype=np.int32)
    num_idx_neg = np.array(num_idx_neg,dtype=np.int32)
    item_input_pos = np.array(item_input_pos, dtype=np.int32)
    item_input_neg = np.array(item_input_neg, dtype=np.int32)
    num_training_instances = len(user_input_pos)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input_pos = user_input_pos[shuffle_index]
    user_input_neg = user_input_neg[shuffle_index]
    num_idx_pos = num_idx_pos[shuffle_index]
    num_idx_neg = num_idx_neg[shuffle_index]
    item_input_pos = item_input_pos[shuffle_index]
    item_input_neg = item_input_neg[shuffle_index]    
    return user_input_pos,user_input_neg, num_idx_pos, num_idx_neg, item_input_pos,item_input_neg

def _get_pointwise_all_likefism_data(dataset,num_negatives):
    user_input,num_idx,item_input,lables = [],[],[],[]
    num_users = dataset.num_users
    num_items = dataset.num_items
    trainMatrix = dataset.trainMatrix
    trainDict = dataset.trainDict
    for u in range(num_users):
        items_by_user = trainDict[u].copy()
        size = len(items_by_user)   
        for i in items_by_user:
            # negative instances
            for _ in range(num_negatives):
                j = np.random.randint(num_items)
                while (u,j) in trainMatrix.keys():
                    j = np.random.randint(num_items)
                user_input.append(items_by_user)
                item_input.append(j)
                num_idx.append(size)
                lables.append(0)
            items_by_user.remove(i)
            user_input.append(items_by_user)
            item_input.append(i)
            num_idx.append(size-1)
            lables.append(1)
    user_input = np.array(user_input)
    num_idx = np.array(num_idx, dtype=np.int32)
    item_input = np.array(item_input, dtype=np.int32)
    lables = np.array(lables, dtype=np.float32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input = user_input[shuffle_index]
    num_idx = num_idx[shuffle_index]
    item_input = item_input[shuffle_index]
    lables = lables[shuffle_index]
    return user_input,num_idx,item_input,lables

def _get_pointwise_all_data(dataset,num_negatives):
    user_input,item_input,lables = [],[],[]
    trainMatrix = dataset.trainMatrix
    num_items= dataset.num_items
    for (u, i) in trainMatrix.keys():
        # positive instance
        user_input.append(u)
        item_input.append(i)
        lables.append(1)
        # negative instance
        for _ in range(num_negatives):
            j = np.random.randint(num_items)
            while (u, j) in trainMatrix.keys():
                j = np.random.randint(num_items)
            user_input.append(u)
            item_input.append(j)
            lables.append(0)
    user_input = np.array(user_input, dtype=np.int32)
    item_input = np.array(item_input, dtype=np.int32)
    lables = np.array(lables, dtype=np.float32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input=user_input[shuffle_index]
    item_input=item_input[shuffle_index]
    lables = lables[shuffle_index]
    return user_input,item_input,lables

def _get_pointwise_all_highorder_data(dataset,high_order,num_negatives):
    user_input, item_input,item_input_recents,lables = [], [], [], []
    trainMatrix = dataset.trainMatrix
    num_items = dataset.num_items
    num_users = dataset.num_users
    trainDict = dataset.trainDict
    for u in range(num_users):
        items_by_user = trainDict[u]
        for idx in range(high_order,len(items_by_user)):
            i = items_by_user[idx] # item id 
            # positive instance
            user_input.append(u)
            item_input.append(i)
            item_input_recent = []
            for t in range(1,high_order+1):
                item_input_recent.append(items_by_user[idx-t])
            item_input_recents.append(item_input_recent)
            lables.append(1)
            for _ in range(num_negatives):
                j = np.random.randint(num_items)
                while (u, j) in trainMatrix.keys():
                    j = np.random.randint(num_items)
                user_input.append(u)
                item_input.append(j)
                item_input_recent = []
                for t in range(1,high_order+1):
                    item_input_recent.append(items_by_user[idx-t])
                item_input_recents.append(item_input_recent)
                lables.append(0)        
    user_input = np.array(user_input, dtype=np.int32)
    item_input = np.array(item_input, dtype=np.int32)
    item_input_recents = np.array(item_input_recents)
    lables = np.array(lables, dtype=np.int32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input = user_input[shuffle_index]
    item_input = item_input[shuffle_index]
    item_input_recents = item_input_recents[shuffle_index]
    lables = lables[shuffle_index]    
    return user_input, item_input, item_input_recents, lables 

def _get_pointwise_all_firstorder_data(dataset,num_negatives):
    trainMatrix = dataset.trainMatrix
    trainDict = dataset.trainDict 
    user_input,item_input,item_input_recent,lables = [],[],[],[]
    num_items = dataset.num_items
    num_users = dataset.num_users
    for u in range(num_users):
        items_by_user = trainDict[u]
        for idx in range(1,len(items_by_user)):
            i = items_by_user[idx] # item id 
            user_input.append(u)
            item_input.append(i)
            item_input_recent.append(items_by_user[idx-1])
            lables.append(1)
            # negative instance
            for _ in range(num_negatives):
                j = np.random.randint(num_items)
                while (u, j) in trainMatrix.keys():
                    j = np.random.randint(num_items)
                user_input.append(u)
                item_input.append(j)
                item_input_recent.append(items_by_user[idx-1])
                lables.append(0)
    user_input = np.array(user_input, dtype=np.int32)
    item_input = np.array(item_input, dtype=np.int32)
    item_input_recent = np.array(item_input_recent, dtype=np.int32)
    lables = np.array(lables, dtype=np.float32)
    num_training_instances = len(user_input)
    shuffle_index = np.arange(num_training_instances,dtype=np.int32)
    np.random.shuffle(shuffle_index)
    user_input = user_input[shuffle_index]
    item_input = item_input[shuffle_index]
    item_input_recent = item_input_recent[shuffle_index]
    lables = lables[shuffle_index]
    return user_input,item_input,item_input_recent,lables
   
def _get_pairwise_batch_likefism_data(user_input_pos,user_input_neg,num_items, num_idx_pos, num_idx_neg, item_input_pos,item_input_neg,num_batch,batch_size):
    num_training_instances = len(user_input_pos)
    id_start = num_batch * batch_size
    id_end = (num_batch + 1) * batch_size
    if id_end>num_training_instances:
        id_end=num_training_instances
    bat_idx_pos = num_idx_pos[id_start:id_end]
    bat_idx_neg = num_idx_neg[id_start:id_end]
    max_pos = max(bat_idx_pos)  
    max_neg = max(bat_idx_neg) 
    bat_users_pos = user_input_pos[id_start:id_end].tolist()
    bat_users_neg = user_input_neg[id_start:id_end].tolist()
    for i in range(len(bat_users_pos)):
        bat_users_pos[i] = bat_users_pos[i] + \
        [num_items] * (max_pos - len(bat_users_pos[i]))
        
    for i in range(len(bat_users_neg)):
        bat_users_neg[i] = bat_users_neg[i] + \
        [num_items] * (max_neg - len(bat_users_neg[i]))  
    
    bat_items_pos = item_input_pos[id_start:id_end]
    bat_items_neg = item_input_neg[id_start:id_end]
    return bat_users_pos,bat_users_neg,bat_idx_pos,bat_idx_neg,bat_items_pos,bat_items_neg


def _get_pairwise_batch_seqdata(user_input,item_input_pos,item_input_recent,item_input_neg,num_batch,batch_size):
    num_training_instances = len(user_input)
    id_start = num_batch * batch_size
    id_end = (num_batch + 1) * batch_size
    if id_end>num_training_instances:
        id_end=num_training_instances
    bat_users = user_input[id_start:id_end]
    bat_items_pos = item_input_pos[id_start:id_end]
    bat_item_recent = item_input_recent[id_start:id_end]
    bat_items_neg = item_input_neg[id_start:id_end]
    return bat_users,bat_items_pos,bat_item_recent,bat_items_neg

 
def _get_pointwise_batch_seqdata(user_input,item_input,item_input_recent,lables,num_batch,batch_size):
    num_training_instances = len(user_input)
    id_start = num_batch * batch_size
    id_end = (num_batch + 1) * batch_size
    if id_end>num_training_instances:
        id_end=num_training_instances
    bat_users = user_input[id_start:id_end]
    bat_items = item_input[id_start:id_end]
    bat_item_recent = item_input_recent[id_start:id_end]
    bat_lables = lables[id_start:id_end]
    return bat_users,bat_items,bat_item_recent,bat_lables
 
def _get_pointwise_batch_likefism_data(user_input,num_items,num_idx,item_input,lables,num_batch,batch_size):
    num_training_instances =len(user_input) 
    id_start = num_batch * batch_size
    id_end = (num_batch + 1) * batch_size
    if id_end>num_training_instances:
        id_end=num_training_instances
    bat_users = user_input[id_start:id_end].tolist()
    bat_idx = num_idx[id_start:id_end]
    max_by_user = max(bat_idx)
    for i in range(len(bat_users)):
        bat_users[i] = bat_users[i] + \
        [num_items] * (max_by_user - len(bat_users[i]))
    bat_items = item_input[id_start:id_end]
    bat_lables = lables[id_start:id_end]
    return bat_users,bat_idx,bat_items,bat_lables
   
def _get_pointwise_batch_data(user_input,item_input,lables,num_batch,batch_size):
    num_training_instances =len(user_input) 
    id_start = num_batch * batch_size
    id_end = (num_batch + 1) * batch_size
    if id_end>num_training_instances:
        id_end=num_training_instances
    bat_users = user_input[id_start:id_end]
    bat_items = item_input[id_start:id_end]
    bat_lables = lables[id_start:id_end]
    return bat_users, bat_items,bat_lables

def _get_pairwise_batch_data(user_input,item_input_pos, item_input_neg, num_batch, batch_size):
    num_training_instances =len(user_input) 
    id_start = num_batch * batch_size
    id_end = (num_batch + 1) * batch_size
    if id_end>num_training_instances:
        id_end=num_training_instances
    bat_users = user_input[id_start:id_end]
    bat_items_pos = item_input_pos[id_start:id_end]
    bat_items_neg = item_input_neg[id_start:id_end]
    return bat_users, bat_items_pos,bat_items_neg


'''
Reference: Lei, Zheng, et al. "Spectral Collaborative Filtering." in RecSys2018
@author: wubin
'''
import tensorflow as tf
import numpy as np
from time import time
from neurec.util import data_gen, learner, tool
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties

class SpectralCF(AbstractRecommender):
    properties = [
        "learning_rate",
        "learner",
        "topk",
        "batch_size",
        "num_layers",
        "activation",
        "embedding_size",
        "epochs",
        "reg",
        "loss_function",
        "dropout",
        "verbose"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.batch_size= self.conf["batch_size"]
        self.num_layers = self.conf["num_layers"]
        self.activation = self.conf["activation"]
        self.embedding_size = self.conf["embedding_size"]
        self.num_epochs=self.conf["epochs"]
        self.reg=self.conf["reg"]
        self.loss_function=self.conf["loss_function"]
        self.dropout=self.conf["dropout"]
        self.verbose=self.conf["verbose"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
        self.graph = self.dataset.trainMatrix.toarray()
        self.A = self.adjacient_matrix(self_connection=True)
        self.D = self.degree_matrix()
        self.L = self.laplacian_matrix(normalized=True)
        self.lamda, self.U = np.linalg.eig(self.L)
        self.lamda = np.diag(self.lamda)

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.input_user = tf.placeholder(dtype=tf.int32, shape=[None,])
            self.input_item = tf.placeholder(dtype=tf.int32, shape=[None,])
            self.input_item_neg = tf.placeholder(dtype=tf.int32, shape=[None,])
    def _create_variables(self):
        with tf.name_scope("embedding"):  # The embedding initialization is unknown now
            self.user_embeddings = tf.Variable(
            tf.random_normal([self.num_users, self.embedding_size], mean=0, stddev=0.01, dtype=tf.float32),
            name='user_embeddings')
            self.item_embeddings = tf.Variable(
            tf.random_normal([self.num_items, self.embedding_size], mean=0, stddev=0.01, dtype=tf.float32),
            name='item_embeddings')

        self.filters = []
        for _ in range(self.num_layers):
            self.filters.append(
                tf.Variable(tf.random_normal([self.embedding_size, self.embedding_size],
                    mean=0, stddev=0.01, dtype=tf.float32)))

    def _create_inference(self):
        with tf.name_scope("inference"):
            A_hat = np.dot(self.U, self.U.T) + np.dot(np.dot(self.U, self.lamda), self.U.T)
            #A_hat += np.dot(np.dot(self.U, self.lamda_2), self.U.T)
            A_hat = A_hat.astype(np.float32)

            embeddings = tf.concat([self.user_embeddings, self.item_embeddings], axis=0)
            all_embeddings = [embeddings]
            for k in range(0, self.num_layers):

                embeddings = tf.matmul(A_hat, embeddings)

                #filters = self.filters[k]#tf.squeeze(tf.gather(self.filters, k))
                embeddings = tool.activation_function(self.activation, (tf.matmul(embeddings, self.filters[k])))
                all_embeddings += [embeddings]
            all_embeddings = tf.concat(all_embeddings, 1)
            self.user_new_embeddings, self.item_new_embeddings = tf.split(all_embeddings, [self.num_users, self.num_items], 0)
            self.u_embeddings = tf.nn.embedding_lookup(self.user_new_embeddings, self.input_user)
            self.pos_i_embeddings = tf.nn.embedding_lookup(self.item_new_embeddings, self.input_item)
            self.neg_j_embeddings = tf.nn.embedding_lookup(self.item_new_embeddings, self.input_item_neg)

    def _create_loss(self):
        with tf.name_scope("loss"):
            self.output = tf.reduce_sum(tf.multiply(self.u_embeddings, self.pos_i_embeddings), axis=1)
            output_neg = tf.reduce_sum(tf.multiply(self.u_embeddings, self.neg_j_embeddings), axis=1)
            regularizer = self.reg * ( tf.reduce_sum(tf.square(self.u_embeddings)) \
                + tf.reduce_sum(tf.square(self.pos_i_embeddings)) + tf.reduce_sum(tf.square(self.neg_j_embeddings)))

            self.loss = learner.pairwise_loss(self.loss_function,self.output - output_neg)+ regularizer

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_inference()
        self._create_loss()
        self._create_optimizer()

    def adjacient_matrix(self, self_connection=False):
        A = np.zeros([self.num_users+self.num_items, self.num_users+self.num_items], dtype=np.float32)
        A[:self.num_users, self.num_users:] = self.graph
        A[self.num_users:, :self.num_users] = self.graph.transpose()
        if self_connection == True:
            return np.identity(self.num_users+self.num_items,dtype=np.float32) + A
        return A

    def degree_matrix(self):
        degree = np.sum(self.A, axis=1, keepdims=False)
        #degree = np.diag(degree)
        return degree

    def laplacian_matrix(self, normalized=False):
        if normalized == False:
            return self.D - self.A

        temp = np.dot(np.diag(np.power(self.D, -1)), self.A)
        #temp = np.dot(temp, np.power(self.D, -0.5))
        return np.identity(self.num_users+self.num_items,dtype=np.float32) - temp

    def train_model(self):
        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input_pos, item_input_neg = data_gen._get_pairwise_all_data(self.dataset)

            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                bat_users,bat_items_pos,bat_items_neg =\
                 data_gen._get_pairwise_batch_data(user_input,\
                 item_input_pos, item_input_neg, num_batch, self.batch_size)
                feed_dict = {self.input_user:bat_users,self.input_item:bat_items_pos,\
                            self.input_item_neg:bat_items_neg}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss

            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        users = np.full(len(items), user_id, dtype=np.int32)
        return self.sess.run(self.output, feed_dict={self.input_user: users, self.input_item: items})



'''
Reference: Xiangnan He, et al., Adversarial Personalized Ranking for Recommendation" in SIGIR2018
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner,data_gen
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
from neurec.util.properties import Properties
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

class APR(AbstractRecommender):
    properties = [
        "learning_rate",
        "embedding_size",
        "learner",
        "topk",
        "epochs",
        "eps",
        "adv",
        "adver",
        "adv_epoch",
        "reg",
        "reg_adv",
        "batch_size",
        "verbose",
        "loss_function"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.learner = self.conf["learner"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.eps= self.conf["eps"]
        self.adv = self.conf["adv"]
        self.adver = self.conf["adver"]
        self.adv_epoch = self.conf["adv_epoch"]
        self.reg = self.conf["reg"]
        self.reg_adv = self.conf["reg_adv"]
        self.batch_size= self.conf["batch_size"]
        self.verbose= self.conf["verbose"]
        self.loss_function = self.conf["loss_function"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape=[None,], name="user_input")
            self.item_input_pos = tf.placeholder(tf.int32, shape=[None,], name="item_input_pos")
            self.item_input_neg = tf.placeholder(tf.int32, shape=[None,], name="item_input_neg")

    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.embedding_P = tf.Variable(
                tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                name='embedding_P', dtype=tf.float32)  # (users, embedding_size)
            self.embedding_Q = tf.Variable(
                tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='embedding_Q', dtype=tf.float32)  # (items, embedding_size)

            self.delta_P = tf.Variable(tf.zeros(shape=[self.num_users, self.embedding_size]),
                                       name='delta_P', dtype=tf.float32, trainable=False)  # (users, embedding_size)
            self.delta_Q = tf.Variable(tf.zeros(shape=[self.num_items, self.embedding_size]),
                                       name='delta_Q', dtype=tf.float32, trainable=False)  # (items, embedding_size)

    def _create_inference(self, item_input):
        with tf.name_scope("inference"):
            # embedding look up
            self.embedding_p = tf.nn.embedding_lookup(self.embedding_P, self.user_input)
            self.embedding_q = tf.nn.embedding_lookup(self.embedding_Q, item_input)  # (b, embedding_size)
            return tf.reduce_sum(self.embedding_p * self.embedding_q,1) # (b, embedding_size) * (embedding_size, 1)

    def _create_inference_adv(self, item_input):
        with tf.name_scope("inference_adv"):
            # embedding look up
            self.embedding_p = tf.nn.embedding_lookup(self.embedding_P, self.user_input)
            self.embedding_q = tf.nn.embedding_lookup(self.embedding_Q, item_input)  # (b, embedding_size)
            # add adversarial noise
            self.P_plus_delta = self.embedding_p + tf.nn.embedding_lookup(self.delta_P, self.user_input)
            self.Q_plus_delta = self.embedding_q + tf.nn.embedding_lookup(self.delta_Q, item_input)
            return tf.reduce_sum(self.P_plus_delta * self.Q_plus_delta,1) # (b, embedding_size) * (embedding_size, 1)

    def _create_loss(self):
        with tf.name_scope("loss"):
            # loss for L(Theta)
            self.output = self._create_inference(self.item_input_pos)
            self.output_neg = self._create_inference(self.item_input_neg)
            self.result = self.output - self.output_neg
            # self.loss = tf.reduce_sum(tf.log(1 + tf.exp(-self.result))) # this is numerically unstable
            self.loss = tf.reduce_sum(tf.nn.softplus(-self.result))

            # loss to be omptimized
            self.opt_loss = self.loss + self.reg * (
                        tf.reduce_sum(tf.square(self.embedding_P)) + tf.reduce_sum(tf.square(self.embedding_Q)))

            if self.adver:
                # loss for L(Theta + adv_Delta)
                self.output_adv = self._create_inference_adv(self.item_input_pos)
                self.output_neg_adv = self._create_inference_adv(self.item_input_neg)
                self.result_adv = self.output_adv - self.output_neg_adv
                # self.loss_adv = tf.reduce_sum(tf.log(1 + tf.exp(-self.result_adv)))
                self.loss_adv = tf.reduce_sum(tf.nn.softplus(-self.result_adv))
                self.opt_loss += self.reg_adv * self.loss_adv


    def _create_adversarial(self):
        with tf.name_scope("adversarial"):
            # generate the adversarial weights by random method
            if self.adv == "random":
                # generation
                self.adv_P = tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01)
                self.adv_Q = tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01)

                # normalization and multiply epsilon
                self.update_P = self.delta_P.assign(tf.nn.l2_normalize(self.adv_P, 1) * self.eps)
                self.update_Q = self.delta_Q.assign(tf.nn.l2_normalize(self.adv_Q, 1) * self.eps)

            # generate the adversarial weights by gradient-based method
            elif self.adv == "grad":
                # return the IndexedSlice Data: [(values, indices, dense_shape)]
                # grad_var_P: [grad,var], grad_var_Q: [grad, var]
                self.grad_P, self.grad_Q = tf.gradients(self.loss, [self.embedding_P, self.embedding_Q])

                # convert the IndexedSlice Data to Dense Tensor
                self.grad_P_dense = tf.stop_gradient(self.grad_P)
                self.grad_Q_dense = tf.stop_gradient(self.grad_Q)

                # normalization: new_grad = (grad / |grad|) * eps
                self.update_P = self.delta_P.assign(tf.nn.l2_normalize(self.grad_P_dense, 1) * self.eps)
                self.update_Q = self.delta_Q.assign(tf.nn.l2_normalize(self.grad_Q_dense, 1) * self.eps)

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
        self._create_adversarial()
#---------- training process -------
    def train_model(self):
        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input_pos, item_input_neg = data_gen._get_pairwise_all_data(self.dataset)
            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                bat_users,bat_items_pos,bat_items_neg =\
                 data_gen._get_pairwise_batch_data(user_input,\
                 item_input_pos, item_input_neg, num_batch, self.batch_size)
                feed_dict = {self.user_input:bat_users,self.item_input_pos:bat_items_pos,\
                            self.item_input_neg:bat_items_neg}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        users = np.full(len(items), user_id, dtype=np.int32)
        return self.sess.run(self.output, feed_dict={self.user_input: users, self.item_input_pos: items})



import logging
from neurec.util.singleton import Singleton
from neurec.util import reader
from neurec.data.properties import types

class Properties(metaclass=Singleton):
    """A class to handle property settings."""
    def __init__(self, properties="", section="DEFAULT"):
        """Setups the class with an empty properties dictionary."""
        self.logger = logging.getLogger(__name__)
        self.__section = section
        self.__properties = self.setProperties(properties) if properties else {}

    def setSection(self, name):
        """Sets the section of the property file to read from.

        name -- name of the section
        """
        self.__section = name

    def setProperties(self, path):
        """Reads a properties file to load the property values.

        path -- path to properties file
        """
        self.__properties = reader.file(path)

    def getProperty(self, name):
        """Returns the value for a property."""
        try:
            value = self.__properties[self.__section][name]
        except KeyError:
            raise KeyError('Key ' + str(name) + ' not found in properties. Add to your properties')

        return self.__convertProperty(name, value)

    def getProperties(self, names):
        """Returns a dictionary of values for the properties names.

        names -- list of property names
        """
        values = {}

        for name in names:
            values[name] = self.getProperty(name)

        return values

    def __convertProperty(self, name, value):
        """Casts a property's value to its required type.

        name -- name of property
        value -- value to convert
        """
        try:
            return types[name](value)
        except KeyError:
            raise KeyError("Could not convert property " + str(name) + ". Key not found in types. Add to neurec.data.properties.types")
        except ValueError:
            raise ValueError("Could not covert the value of " + str(name) + '. ' + str(value) + " does not match type set in neurec.data.properties.types")



'''
Reference: Yifan Hu et al., "Collaborative Filtering for Implicit Feedback Datasets" in ICDM 2008.
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
import numpy as np
from time import time
from neurec.evaluation import Evaluate
from neurec.model.AbstractRecommender import AbstractRecommender
import tensorflow as tf
from neurec.util.properties import Properties

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
class WRMF(AbstractRecommender):
    properties = [
        "embedding_size",
        "alpha",
        "topk",
        "epochs",
        "reg_mf",
        "verbose"
    ]

    def __init__(self, sess, dataset):
        super().__init__(**kwds)

        self.embedding_size = self.conf["embedding_size"]
        self.alpha = self.conf["alpha"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.reg_mf = self.conf["reg_mf"]
        self.verbose= self.conf["verbose"]

        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

        self.Cui = np.zeros(shape=[self.num_users, self.num_items], dtype=np.float32)
        self.Pui = np.zeros(shape=[self.num_users, self.num_items], dtype=np.float32)
        for (u, i) in self.dataset.trainMatrix.keys():
            self.Cui[u,i] = self.alpha
            self.Pui[u,i] = 1.0
        self.lambda_eye = self.reg_mf * tf.eye(self.embedding_size)

    def _create_placeholders(self):
        self.user_id = tf.placeholder(tf.int32, [1])
        self.Cu = tf.placeholder(tf.float32, [self.num_items, 1])
        self.Pu = tf.placeholder(tf.float32, [self.num_items, 1])

        self.item_id = tf.placeholder(tf.int32, [1])
        self.Ci = tf.placeholder(tf.float32, [self.num_users, 1])
        self.Pi = tf.placeholder(tf.float32, [self.num_users, 1])

    def _create_variables(self):
        self.user_embeddings = tf.Variable(tf.random_normal([self.num_users, self.embedding_size], stddev=0.01))
        self.item_embeddings = tf.Variable(tf.random_normal([self.num_items, self.embedding_size], stddev=0.01))

    def _create_optimizer(self):
        YTY = tf.matmul(self.item_embeddings, self.item_embeddings, transpose_a=True)
        YTCuIY = tf.matmul(self.item_embeddings, tf.multiply(self.Cu, self.item_embeddings), transpose_a=True)
        YTCupu = tf.matmul(self.item_embeddings, tf.multiply(self.Cu+1, self.Pu), transpose_a=True)
        xu = tf.linalg.solve(YTY + YTCuIY + self.lambda_eye, YTCupu)
        self.update_user = tf.scatter_update(self.user_embeddings, self.user_id, tf.transpose(xu))

        XTX = tf.matmul(self.user_embeddings, self.user_embeddings, transpose_a=True)
        XTCIIX = tf.matmul(self.user_embeddings, tf.multiply(self.Ci, self.user_embeddings), transpose_a=True)
        XTCIpi = tf.matmul(self.user_embeddings, tf.multiply(self.Ci+1, self.Pi), transpose_a=True)
        xi = tf.linalg.solve(XTX + XTCIIX + self.lambda_eye, XTCIpi)
        self.update_item = tf.scatter_update(self.item_embeddings, self.item_id, tf.transpose(xi))

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_optimizer()

        #---------- training process -------
    def train_model(self):
        for epoch in  range(self.num_epochs):
            training_start_time = time()
            self.logger.info('solving for user vectors...')
            for userid in range(self.num_users):
                feed = {self.user_id: [userid],
                        self.Pu: self.Pui[userid].T.reshape([-1,1]),
                        self.Cu: self.Cui[userid].T.reshape([-1,1])}
                self.sess.run(self.update_user, feed_dict=feed)

            self.logger.info('solving for item vectors...')
            for itemid in range(self.num_items):
                feed = {self.item_id: [itemid],
                        self.Pi: self.Pui[:,itemid].reshape([-1,1]),
                        self.Ci: self.Cui[:,itemid].reshape([-1,1])}
                self.sess.run(self.update_item, feed_dict=feed)

            self.logger.info ('iteration %i finished in %f seconds' % (epoch + 1, time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)

    def predict(self, user_id, items):
        user_embeddings, item_embeddings = self.sess.run([self.user_embeddings, self.item_embeddings])
        user_embedding = user_embeddings[user_id]
        item_embedding = item_embeddings[items]
        predictions = user_embedding.dot(item_embedding.T)
        return predictions



def hit(ranklist,taget_items):
    count = 0
    for item in ranklist:
        if item in taget_items:
            count += 1
    return count

def getRec(ranklist,taget_items):
    
    return hit(ranklist, taget_items)/len(taget_items)


import numpy

def to_list(string):
    """Returns a list from a string.

    string -- string in format [0.5,0.5]|type
    """
    pipe_split = string.split("|")

    try:
        cast = pipe_split[1]
        array = pipe_split[0]
        array_values = array[1:-1].split(",")
    except IndexError:
        raise IndexError("list index out of range for " + str(string) + ". Make sure the list is the form [0,0]|type")
        
    return numpy.array(array_values, dtype=cast)

def to_bool(string):
    """Returns a boolean from a string.

    string -- string in format yes|no|true|false
    """
    if string.lower() in ("yes", "y", "true", "1"):
        return True
    elif string.lower() in ("no", "n", "false", "0"):
        return False

    raise ValueError(str(string) + ' not in ["yes", "y", "true", "1", "no", "n", "false", "0"].')



from neurec.util.helpers import to_list, to_bool

types = {
    "data.input.path": str,
    "data.input.dataset": str,
    "data.splitter": str,
    "data.column.format": str,
    "data.convert.separator": eval,
    "data.convert.binarize.threshold": float,
    "recommender": str,
    "rec.evaluate.neg": int,
    "data.splitterratio": to_list,
    "rec.number.thread": int,
    "topk": int,
    "epochs": int,
    "batch_size": int,
    "layers": to_list,
    "embedding_size": int,
    "reg": float,
    "reg_adv": int,
    "reg_g": float,
    "reg_d": float,
    "learning_rate": float,
    "learner": str,
    "adv_epoch": int,
    "adv": str,
    "eps": float,
    "adver": int,
    "verbose": int,
    "loss_function": str,
    "reg_mlp": float,
    "ispairwise": to_bool,
    "num_neg": int,
    "hidden_neuron": int,
    "h_act": str,
    "g_act": str,
    "corruption_level": float,
    "mode": str,
    "lr_g": float,
    "lr_d": float,
    "batchsize_g": int,
    "batchsize_d": int,
    "opt_g": str,
    "opt_d": str,
    "hiddenlayer_g": to_list,
    "hiddenlayer_d": to_list,
    "step_g": int,
    "step_d": int,
    "zr_ratio": float,
    "zp_ratio": float,
    "zr_coefficient": float,
    "regs": to_list,
    "keep": float,
    "net_channel": to_list,
    "lr_embed": float,
    "lr_net": float,
    "pretrain": int,
    "batch_choice": str,
    "batch_norm": int,
    "weight_size": int,
    "regw": to_list,
    "alpha": float,
    "beta": float,
    "activation": str,
    "algorithm": int,
    "reg_mf": float,
    "pre_agg": str,
    "session_agg": str,
    "high_order": int,
    "factors_num": int,
    "lr": float,
    "g_reg": float,
    "d_reg": eval,
    "g_epoch": int,
    "d_epoch": int,
    "d_tau": float,
    "pretrain_file": str,
    "f_act": str,
    "margin": float,
    "p_dim": to_list,
    "anneal_cap": float,
    "total_anneal_steps": int,
    "data_alpha": float,
    "mf_pretrain": str,
    "mlp_pretrain": str,
    "layer_size": to_list,
    "node_dropout_flag": to_bool,
    "adj_type": str,
    "alg_type": str,
    "socialpath": str,
    "num_epochs": int,
    "num_layers": int,
    "dropout": float
}



import tensorflow as tf
import time
def activation_function(act,act_input):
        act_func = None
        if act == "sigmoid":
            act_func = tf.nn.sigmoid(act_input)
        elif act == "tanh":
            act_func = tf.nn.tanh(act_input)
            
        elif act == "relu":
            act_func = tf.nn.relu(act_input)
        
        elif act == "elu":
            act_func = tf.nn.elu(act_input)
           
        elif act == "identity":
            act_func = tf.identity(act_input)
        else:
            raise NotImplementedError("ERROR")
        return act_func  
def getlocaltime():
    date = time.strftime('%y-%m-%d', time.localtime())
    current_time = time.strftime('%H:%M:%S', time.localtime())      


def getMRR(rank_list, target_items):
    for i in range(len(rank_list)):
        item_idx = rank_list[i] 
        if item_idx in target_items:
            return 1/(i+1)
        else:
            return 0


import scipy.sparse as sp
import numpy as np
from copy import deepcopy
from neurec.util import reader
import logging

class LeaveOneOutDataSplitter(object):
    def __init__(self, path, dataset_name, data_format, separator, threshold):
        self.path =path
        self.dataset_name = dataset_name
        self.data_format = data_format
        self.separator = separator
        self.threshold = threshold
        self.logger = logging.getLogger("neurec.data.LeaveOneOutDataSplitter.LeaveOneOutDataSplitter")
    def load_data_by_user_time(self):
        self.logger.info("Loading interaction records from %s "%(self.path))
        pos_per_user = {}
        num_ratings=0
        num_items=0
        num_users=0
        #user/item {raw id, inner id} map
        userids = {}
        itemids = {}
        # inverse views of userIds, itemIds,
        idusers = {}
        iditems={}

        data = reader.lines(self.path + '/' + self.dataset_name)
        for line in data:
            if self.data_format == "UIRT":
                useridx, itemidx,rating,time= line.strip().split(self.separator)
                if float(rating) < self.threshold:
                    continue
            elif self.data_format == "UIT":
                useridx, itemidx,time= line.strip().split(self.separator)
                rating = 1
            elif self.data_format == "UIR":
                useridx, itemidx,rating = line.strip().split(self.separator)
                if float(rating) < self.threshold:
                    continue
            elif self.data_format == "UI":
                useridx, itemidx = line.strip().split(self.separator)
                rating = 1

            else:
                print("please choose a correct data format. ")

            num_ratings+=1
            if  itemidx not in itemids:
                iditems[num_items]=itemidx
                itemids[itemidx] = num_items
                num_items+=1

            if useridx not in userids:
                idusers[num_users]=useridx
                userids[useridx]=num_users
                num_users+=1
                pos_per_user[userids[useridx]]=[]
            if  self.data_format == "UIRT" or self.data_format == "UIT":
                pos_per_user[userids[useridx]].append((itemids[itemidx],rating,int(float(time))))

            else:
                pos_per_user[userids[useridx]].append((itemids[itemidx],rating,1))

        if  self.data_format == "UIRT" or self.data_format == "UIT":
            for u in np.arange(num_users):
                pos_per_user[u]=sorted(pos_per_user[u], key=lambda d: d[2])
        self.logger.info("\"num_users\": %d,\"num_items\":%d, \"num_ratings\":%d\n"%(num_users,num_items,num_ratings))
        userseq = deepcopy(pos_per_user)
        train_dict = {}
        time_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
        train_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
        test_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
        for u in np.arange(num_users):
            if len(pos_per_user[u])<2:
                test_item=-1
                continue
            test_item=pos_per_user[u][-1]
            pos_per_user[u].pop()

            test_matrix[u,test_item[0]] = test_item[1]
            time_matrix[u,test_item[0]] = test_item[2]
            items = []
            for enlement in pos_per_user[u]:
                items.append(enlement[0])
                train_matrix[u,enlement[0]]=enlement[1]
                time_matrix[u,enlement[0]] = enlement[2]
            train_dict[u]=items

            for u in np.arange(num_users):
                pos_per_user[u]=sorted(pos_per_user[u], key=lambda d: d[2])
            self.logger.info("\"num_users\": %d,\"num_items\":%d, \"num_ratings\":%d\n"%(num_users,num_items,num_ratings))
            userseq = deepcopy(pos_per_user)
            train_dict = {}
            train_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
            test_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
            time_matrix = sp.dok_matrix((num_users, num_items), dtype=np.float32)
            for u in np.arange(num_users):
                if len(pos_per_user[u])<3:
                    test_item=-1
                    continue

                test_item=pos_per_user[u][-1]
                pos_per_user[u].pop()

                test_matrix[u,test_item[0]] = test_item[1]
                time_matrix[u,test_item[0]] = test_item[2]
                items = []
                for enlement in pos_per_user[u]:
                    items.append(enlement[0])
                    train_matrix[u,enlement[0]]=enlement[1]
                    time_matrix[u,enlement[0]] = enlement[2]
                train_dict[u]=items
        return train_matrix,train_dict,test_matrix,userseq,userids,itemids,time_matrix



'''
Reference: Dong-Kyu Chae, et al., "CFGAN: A Generic Collaborative Filtering Framework
based on Generative Adversarial Networks." in CIKM2018
@author: Zhongchuan Sun
'''
# ZP: Hybrid of zero-reconstruction regularization and partial-masking

from neurec.model.AbstractRecommender import AbstractRecommender
import numpy as np
import tensorflow as tf
from scipy.sparse import csr_matrix
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

def csr_to_user_dict(sparse_matrix_data):
    """convert a scipy.sparse.csr_matrix to a dict,
    where the key is row number, and value is the
    non-empty index in each row.
    """
    idx_value_dict = {}
    for idx, value in enumerate(sparse_matrix_data):
        if any(value.indices):
            idx_value_dict[idx] = value.indices
    return idx_value_dict


def random_choice(a, size=None, replace=True, p=None, exclusion=None):
    if exclusion is not None:
        if p is None:
            p = np.ones_like(a)
        else:
            p = np.array(p, copy=True)
        p = np.ndarray.flatten(p)
        p[exclusion] = 0
        p = p / np.sum(p)
    sample = np.random.choice(a, size=size, replace=replace, p=p)
    return sample


class CFGAN(AbstractRecommender):
    properties = [
        "epochs",
        "topk",
        "mode",
        "reg_g",
        "reg_d",
        "lr_g",
        "lr_d",
        "batchsize_g",
        "batchsize_d",
        "opt_g",
        "opt_d",
        "hiddenlayer_g",
        "hiddenlayer_d",
        "step_g",
        "step_d",
        "zr_ratio",
        "zp_ratio",
        "zr_coefficient",
        "verbose"
    ]

    def __init__(self, sess, dataset):
        super().__init__(**kwds)

        self.epochs = self.conf["epochs"]
        self.topK = self.conf["topk"]
        self.mode = self.conf["mode"]
        self.reg_G = self.conf["reg_g"]
        self.reg_D = self.conf["reg_d"]
        self.lr_G = self.conf["lr_g"]
        self.lr_D = self.conf["lr_d"]
        self.batchSize_G = self.conf["batchsize_g"]
        self.batchSize_D = self.conf["batchsize_d"]

        self.opt_G = self.conf["opt_g"]
        self.opt_D = self.conf["opt_d"]
        self.hiddenLayer_G = self.conf["hiddenlayer_g"]
        self.hiddenLayer_D = self.conf["hiddenlayer_d"]
        self.step_G = self.conf["step_g"]
        self.step_D = self.conf["step_d"]

        self.ZR_ratio = self.conf["zr_ratio"]
        self.ZP_ratio = self.conf["zp_ratio"]
        self.ZR_coefficient = self.conf["zr_coefficient"]
        self.verbose= self.conf["verbose"]

        train_matrix = dataset.trainMatrix.tocsr()
        self.train_matrix = train_matrix.copy()
        if self.mode == "itemBased":
            self.train_matrix = self.train_matrix.transpose(copy=True).tocsr()

        self.num_users, self.num_items = self.train_matrix.shape
        self.user_pos_train = csr_to_user_dict(self.train_matrix)
        self.all_items = np.arange(self.num_items)
        self.loss_function = "None"

        # self._build_model()
        # self.sess.run(tf.global_variables_initializer())

    def build_graph(self):
        self._create_layer()

        # generator
        self.condition = tf.placeholder(tf.float32, [None, self.num_items])
        self.G_ZR_dims = tf.placeholder(tf.float32, [None, self.num_items])
        self.G_output = self.gen(self.condition)
        self.G_ZR_loss = tf.reduce_mean(tf.reduce_sum(tf.square(self.G_output - 0) * self.G_ZR_dims, 1, keepdims=True))

        # discriminator
        self.mask = tf.placeholder(tf.float32, [None, self.num_items])  # purchased = 1, otherwise 0
        fakeData = self.G_output * self.mask
        fakeData = tf.concat([self.condition, fakeData], 1)

        self.realData = tf.placeholder(tf.float32, [None, self.num_items])
        realData = tf.concat([self.condition, self.realData], 1)

        D_fake = self.dis(fakeData)
        D_real = self.dis(realData)

        g_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope='gen')
        d_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, scope='dis')

        # define loss & optimizer for G
        g_loss = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=D_fake, labels=tf.ones_like(D_fake)))
        g_loss = g_loss + self.reg_G*self._l2loss(g_vars)
        g_loss = g_loss + self.ZR_coefficient*self.G_ZR_loss

        if self.opt_G == 'sgd':
            self.trainer_G = tf.train.GradientDescentOptimizer(self.lr_G).minimize(g_loss, var_list=g_vars)
        elif self.opt_G == 'adam':
            self.trainer_G = tf.train.AdamOptimizer(self.lr_G).minimize(g_loss, var_list=g_vars)

        # define loss & optimizer for D
        d_loss_real = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=D_real, labels=tf.ones_like(D_real)))
        d_loss_fake = tf.reduce_mean(tf.nn.sigmoid_cross_entropy_with_logits(logits=D_fake, labels=tf.zeros_like(D_fake)))
        d_loss = d_loss_real + d_loss_fake + self.reg_D*self._l2loss(d_vars)

        if self.opt_D == 'sgd':
            self.trainer_D = tf.train.GradientDescentOptimizer(self.lr_D).minimize(d_loss, var_list=d_vars)
        elif self.opt_D == 'adam':
            self.trainer_D = tf.train.AdamOptimizer(self.lr_D).minimize(d_loss, var_list=d_vars)

    def _l2loss(self, var):
        l2loss = 0
        for v in var:
            l2loss += tf.nn.l2_loss(v)
        return l2loss

    def _create_layer(self):
        # Generator's layers
        self.gen_layers = []
        xavier_init = tf.contrib.layers.xavier_initializer()
        # stacked hidden layers
        for i, unit in enumerate(self.hiddenLayer_G):
            hidden_layer = tf.layers.Dense(unit, activation=tf.sigmoid,
                                           kernel_initializer=xavier_init, name="gen_h%d"%i)
            self.gen_layers.append(hidden_layer)

        # hidden -> output
        output_layer = tf.layers.Dense(self.num_items, activation=tf.identity,
                                       kernel_initializer=xavier_init, name="gen_out")
        self.gen_layers.append(output_layer)

        # Discriminator's layers
        self.dis_layers = []
        # stacked hidden layers
        for i, unit in enumerate(self.hiddenLayer_D):
            hidden_layer = tf.layers.Dense(unit, activation=tf.sigmoid,
                                           kernel_initializer=xavier_init, name="dis_h%d"%i)
            self.dis_layers.append(hidden_layer)

        # hidden -> output
        output_layer = tf.layers.Dense(1, activation=tf.identity,
                                       kernel_initializer=xavier_init, name="dis_out")
        self.dis_layers.append(output_layer)

    def gen(self, input):
        for layer in self.gen_layers:
            input = layer.apply(input)
        return input

    def dis(self, input):
        for layer in self.dis_layers:
            input = layer.apply(input)
        return input

    def get_train_data(self):
        train_matrix = self.train_matrix.copy()
        ZR_matrix = self.train_matrix.todense()
        PM_matrix = self.train_matrix.todense()

        for u, pos_items in self.user_pos_train.items():
            num = int((self.num_items-len(pos_items)) * self.ZR_ratio)
            sample = random_choice(self.all_items, size=num, replace=False, exclusion=pos_items).tolist()
            ZR_matrix[u, sample] = 1

            num = int((self.num_items-len(pos_items)) * self.ZP_ratio)
            sample = random_choice(self.all_items, size=num, replace=False, exclusion=pos_items).tolist()
            PM_matrix[u, sample] = 1
        return train_matrix, csr_matrix(ZR_matrix), csr_matrix(PM_matrix)

    def train_model(self):
        gen_batch_index = np.arange(self.num_users)
        np.random.shuffle(gen_batch_index)
        dis_batch_index = np.arange(self.num_users)
        np.random.shuffle(dis_batch_index)

        totalEpochs = self.epochs
        totalEpochs = int(totalEpochs / self.step_G)
        for epoch in range(totalEpochs):
            train_matrix, ZR_matrix, PM_matrix = self.get_train_data()
            # training discriminator
            for d_epoch in range(self.step_D):
                for idx in np.arange(0, self.num_users, step=self.batchSize_D):
                    idx = dis_batch_index[idx:idx + self.batchSize_D]
                    train_data = train_matrix[idx].toarray()
                    train_mask = PM_matrix[idx].toarray()
                    feed = {self.realData: train_data, self.mask: train_mask, self.condition: train_data}
                    self.sess.run(self.trainer_D, feed_dict=feed)

            # training generator
            for g_epoch in range(self.step_G):
                for idx in np.arange(0, self.num_users, step=self.batchSize_G):
                    idx = dis_batch_index[idx:idx + self.batchSize_G]
                    train_data = train_matrix[idx].toarray()
                    train_z_mask = ZR_matrix[idx].toarray()
                    train_p_mask = PM_matrix[idx].toarray()
                    feed = {self.realData: train_data, self.condition: train_data,
                            self.mask: train_p_mask, self.G_ZR_dims: train_z_mask}
                    self.sess.run(self.trainer_G, feed_dict=feed)
            if epoch %self.verbose == 0:
                self.eval_rating_matrix()
                Evaluate.test_model(self, self.dataset)


    def eval_rating_matrix(self):
        allRatings = self.sess.run(self.G_output, feed_dict={self.condition: self.train_matrix.toarray()})
        if self.mode == "itemBased":
            allRatings = np.transpose(allRatings)
        self.allRatings_for_test = allRatings

    def predict(self, user_id, items):
        return self.allRatings_for_test[user_id, items]



def getAUC(predictions,K):
    auc = []
    neg_predict, pos_predict = predictions[:-1], predictions[-1]
    position = (neg_predict >= pos_predict).sum()
    for _ in range(1, K + 1):
        # formula: [#(Xui>Xuj) / #(Items)] = [1 - #(Xui<=Xuj) / #(Items)]
        auc.append(1 - (position / len(neg_predict))) 
        
    return auc


import setuptools

setuptools.setup(
    name="neurec",
    version="1.0.0",
    author="Jonathan Staniforth",
    author_email="jonathanstaniforth@gmail.com",
    description="An open source neural recommender library",
    long_description="NeuRec is a comprehensive and flexible Python library for recommendation models that includes a large range of state-of-the-art neural recommender models. This library aims to solve general and sequential (i.e. next-item) recommendation tasks, using the Tensorflow library to provide 26 models out of the box. NeuRec is open source and available under the MIT license. View the GitHub repository for more information: https://github.com/wubinzzu/NeuRec.git",
    url="https://github.com/wubinzzu/NeuRec.git",
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.5'
)



from neurec.model.item_ranking.MF import MF
from neurec.model.item_ranking.FISM import FISM
from neurec.model.item_ranking.NAIS import NAIS
from neurec.model.item_ranking.MLP import MLP
from neurec.model.item_ranking.NeuMF import NeuMF
from neurec.model.item_ranking.APR import APR
from neurec.model.item_ranking.DMF import DMF
from neurec.model.item_ranking.ConvNCF import ConvNCF
from neurec.model.item_ranking.CDAE import CDAE
from neurec.model.item_ranking.CFGAN import CFGAN
from neurec.model.item_ranking.DAE import DAE
from neurec.model.seq_ranking.NPE import NPE
from neurec.model.item_ranking.IRGAN import IRGAN
from neurec.model.item_ranking.MultiDAE import MultiDAE
from neurec.model.item_ranking.MultiVAE import MultiVAE
from neurec.model.item_ranking.JCA import JCA
from neurec.model.item_ranking.CFGAN import CFGAN
from neurec.model.item_ranking.SBPR import SBPR
from neurec.model.item_ranking.WRMF import WRMF
from neurec.model.item_ranking.SpectralCF import SpectralCF
from neurec.model.item_ranking.DeepICF import DeepICF
from neurec.model.item_ranking.NGCF import NGCF
from neurec.model.seq_ranking.FPMC import FPMC
from neurec.model.seq_ranking.FPMCplus import FPMCplus
from neurec.model.seq_ranking.HRM import HRM
from neurec.model.seq_ranking.TransRec import TransRec

models = {
    "apr": APR,
    "cdae": CDAE,
    "cfgan": CFGAN,
    "convncf": ConvNCF,
    "dae": DAE,
    "deepicf":DeepICF,
    "dmf": DMF,
    "fism": FISM,
    "irgan": IRGAN,
    "jca": JCA,
    "mf": MF,
    "mlp": MLP,
    "multidae": MultiDAE,
    "multivae": MultiVAE,
    "nais": NAIS,
    "neumf": NeuMF,
    "ngcf": NGCF,
    "sbpr": SBPR,
    "spectralcf": SpectralCF,
    "wrmf": WRMF,
    "fpmc": FPMC,
    "fpmcplus": FPMCplus,
    "hrm": HRM,
    "npe": NPE,
    "transrec": TransRec
}



'''
Reference: Ruining He et al., "Translation-based Recommendation." in RecSys 2017
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner, data_gen
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

class TransRec(AbstractRecommender):
    properties = [
        "learning_rate",
        "embedding_size",
        "learner",
        "loss_function",
        "ispairwise",
        "topk",
        "epochs",
        "reg_mf",
        "batch_size",
        "verbose",
        "num_neg"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.learner = self.conf["learner"]
        self.loss_function = self.conf["loss_function"]
        self.ispairwise = self.conf["ispairwise"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.reg_mf = self.conf["reg_mf"]
        self.batch_size= self.conf["batch_size"]
        self.verbose= self.conf["verbose"]
        self.num_negatives= self.conf["num_neg"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input = tf.placeholder(tf.int32, shape = [None,], name = "item_input_pos")
            self.item_input_recent = tf.placeholder(tf.int32, shape = [None,], name = "item_input_recent")
            if self.ispairwise == True:
                self.item_input_neg = tf.placeholder(tf.int32, shape = [None,], name = "item_input_neg")
            else :
                self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")
    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.user_embeddings = tf.Variable(tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                name='user_embeddings', dtype=tf.float32)  #(users, embedding_size)
            self.item_embeddings = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='item_embeddings', dtype=tf.float32)  #(items, embedding_size)
            self.item_biases = tf.Variable(tf.truncated_normal(shape=[self.num_items], mean=0.0, stddev=0.01),
                name='item_biases', dtype=tf.float32)  #(items)
            self.global_embedding = tf.Variable(tf.truncated_normal(shape=[1,self.embedding_size], mean=0.0, stddev=0.01),
                name='global_embedding', dtype=tf.float32)

    def _create_inference(self, item_input):
        with tf.name_scope("inference"):
            # embedding look up
            user_embedding = tf.nn.embedding_lookup(self.user_embeddings, self.user_input)
            batch_size =  tf.shape(user_embedding)[0]
            item_embedding_recent = tf.nn.embedding_lookup(self.item_embeddings, self.item_input_recent)
            item_embedding = tf.nn.embedding_lookup(self.item_embeddings, item_input)
            item_bias = tf.nn.embedding_lookup(self.item_biases, item_input)
            predict_vector = user_embedding + tf.tile(self.global_embedding, tf.stack([batch_size,1]))+item_embedding_recent -item_embedding
            predict = item_bias-tf.reduce_sum(tf.square(predict_vector),1)
            return user_embedding, item_embedding_recent,item_embedding,item_bias,predict

    def _create_loss(self):
        with tf.name_scope("loss"):
            # loss for L(Theta)
            p1,r1,q1,b1,self.output = self._create_inference(self.item_input)
            if self.ispairwise == True:
                _,_,q2,b2,output_neg = self._create_inference(self.item_input_neg)
                self.result = self.output - output_neg
                self.loss = learner.pairwise_loss(self.loss_function,self.result) + self.reg_mf * ( tf.reduce_sum(tf.square(p1)) \
                +tf.reduce_sum(tf.square(r1)) + tf.reduce_sum(tf.square(q2)) + tf.reduce_sum(tf.square(q1))+tf.reduce_sum(tf.square(b1)) + tf.reduce_sum(tf.square(b2)) + tf.reduce_sum(tf.square(self.global_embedding)))
            else :
                self.loss = learner.pointwise_loss(self.loss_function,self.lables,self.output) + self.reg_mf * (tf.reduce_sum(tf.square(p1)) \
                +tf.reduce_sum(tf.square(r1))+ tf.reduce_sum(tf.square(q1))+ tf.reduce_sum(tf.square(b1))+ tf.reduce_sum(tf.square(self.global_embedding)))

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
#---------- training process -------
    def train_model(self):

        for epoch in  range(self.num_epochs):
            # Generate training instances
            if self.ispairwise == True:
                user_input, item_input_pos, item_input_recents, item_input_neg = \
                data_gen._get_pairwise_all_firstorder_data(self.dataset)
            else :
                user_input, item_input,item_input_recents, lables = data_gen._get_pointwise_all_firstorder_data(self.dataset,self.num_negatives)


            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                if self.ispairwise == True:
                    bat_users, bat_items_pos, bat_items_recents,bat_items_neg  = \
                    data_gen._get_pairwise_batch_seqdata(user_input, item_input_pos, \
                    item_input_recents, item_input_neg, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users,self.item_input:bat_items_pos,\
                                self.item_input_recent:bat_items_recents,self.item_input_neg:bat_items_neg}
                else :
                    bat_users, bat_items, bat_items_recents, bat_lables =\
                    data_gen._get_pointwise_batch_seqdata(user_input, \
                    item_input,item_input_recents, lables, num_batch, self.batch_size)
                    feed_dict = {self.user_input:bat_users, self.item_input:bat_items,
                                 self.item_input_recent:bat_items_recents,self.lables:bat_lables}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss

            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)
    def predict(self, user_id, items):
        cand_items = self.dataset.trainDict[user_id]
        item_recent = np.full(len(items), cand_items[-1], dtype='int32')

        users = np.full(len(items), user_id, dtype='int32')
        return self.sess.run((self.output), feed_dict={self.user_input: users,\
                                        self.item_input_recent:item_recent, self.item_input: items})



'''
Reference: Jun Wang, et al., "IRGAN: A Minimax Game for Unifying Generative and
Discriminative Information Retrieval Models." SIGIR 2017.
@author: Zhongchuan Sun
'''
from neurec.model.AbstractRecommender import AbstractRecommender
import tensorflow as tf
import pickle
import numpy as np
from concurrent.futures import ThreadPoolExecutor
from neurec.util import data_gen, reader
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

class GEN(object):
    def __init__(self, itemNum, userNum, emb_dim, lamda, param=None, initdelta=0.05, learning_rate=0.05):
        self.itemNum = itemNum
        self.userNum = userNum
        self.emb_dim = emb_dim
        self.lamda = lamda  # regularization parameters
        self.param = param
        self.initdelta = initdelta
        self.learning_rate = learning_rate
        self.g_params = []

        with tf.variable_scope('generator'):
            if self.param == None:
                self.user_embeddings = tf.Variable(
                    tf.random_uniform([self.userNum, self.emb_dim], minval=-self.initdelta, maxval=self.initdelta,
                                      dtype=tf.float32))
                self.item_embeddings = tf.Variable(
                    tf.random_uniform([self.itemNum, self.emb_dim], minval=-self.initdelta, maxval=self.initdelta,
                                      dtype=tf.float32))
                self.item_bias = tf.Variable(tf.zeros([self.itemNum]))
            else:
                self.user_embeddings = tf.Variable(self.param[0])
                self.item_embeddings = tf.Variable(self.param[1])
                self.item_bias = tf.Variable(param[2])

            self.g_params = [self.user_embeddings, self.item_embeddings, self.item_bias]

        self.u = tf.placeholder(tf.int32)
        self.i = tf.placeholder(tf.int32)
        self.reward = tf.placeholder(tf.float32)

        self.u_embedding = tf.nn.embedding_lookup(self.user_embeddings, self.u)
        self.i_embedding = tf.nn.embedding_lookup(self.item_embeddings, self.i)
        self.i_bias = tf.gather(self.item_bias, self.i)

        self.all_logits = tf.reduce_sum(tf.multiply(self.u_embedding, self.item_embeddings), 1) + self.item_bias
        self.i_prob = tf.gather(
            tf.reshape(tf.nn.softmax(tf.reshape(self.all_logits, [1, -1])), [-1]),
            self.i)

        self.gan_loss = -tf.reduce_mean(tf.log(self.i_prob) * self.reward) + self.lamda * (
            tf.nn.l2_loss(self.u_embedding) + tf.nn.l2_loss(self.i_embedding) + tf.nn.l2_loss(self.i_bias))

        g_opt = tf.train.GradientDescentOptimizer(self.learning_rate)
        self.gan_updates = g_opt.minimize(self.gan_loss, var_list=self.g_params)

        # for test stage, self.u: [batch_size]
        self.all_rating = tf.matmul(self.u_embedding, self.item_embeddings, transpose_a=False,
                                    transpose_b=True) + self.item_bias


class DIS(object):
    def __init__(self, itemNum, userNum, emb_dim, lamda, param=None, initdelta=0.05, learning_rate=0.05):
        self.itemNum = itemNum
        self.userNum = userNum
        self.emb_dim = emb_dim
        self.lamda = lamda  # regularization parameters
        self.param = param
        self.initdelta = initdelta
        self.learning_rate = learning_rate
        self.d_params = []

        with tf.variable_scope('discriminator'):
            if self.param == None:
                self.user_embeddings = tf.Variable(
                    tf.random_uniform([self.userNum, self.emb_dim], minval=-self.initdelta, maxval=self.initdelta,
                                      dtype=tf.float32))
                self.item_embeddings = tf.Variable(
                    tf.random_uniform([self.itemNum, self.emb_dim], minval=-self.initdelta, maxval=self.initdelta,
                                      dtype=tf.float32))
                self.item_bias = tf.Variable(tf.zeros([self.itemNum]))
            else:
                self.user_embeddings = tf.Variable(self.param[0])
                self.item_embeddings = tf.Variable(self.param[1])
                self.item_bias = tf.Variable(self.param[2])

        self.d_params = [self.user_embeddings, self.item_embeddings, self.item_bias]

        # placeholder definition
        self.u = tf.placeholder(tf.int32)
        self.i = tf.placeholder(tf.int32)
        self.label = tf.placeholder(tf.float32)

        self.u_embedding = tf.nn.embedding_lookup(self.user_embeddings, self.u)
        self.i_embedding = tf.nn.embedding_lookup(self.item_embeddings, self.i)
        self.i_bias = tf.gather(self.item_bias, self.i)

        self.pre_logits = tf.reduce_sum(tf.multiply(self.u_embedding, self.i_embedding), 1) + self.i_bias
        self.pre_loss = tf.nn.sigmoid_cross_entropy_with_logits(labels=self.label,
                                                                logits=self.pre_logits) + self.lamda * (
            tf.nn.l2_loss(self.u_embedding) + tf.nn.l2_loss(self.i_embedding) + tf.nn.l2_loss(self.i_bias)
        )

        d_opt = tf.train.GradientDescentOptimizer(self.learning_rate)
        self.d_updates = d_opt.minimize(self.pre_loss, var_list=self.d_params)

        self.reward_logits = tf.reduce_sum(tf.multiply(self.u_embedding, self.i_embedding),
                                           1) + self.i_bias
        self.reward = 2 * (tf.sigmoid(self.reward_logits) - 0.5)

        # for test stage, self.u: [batch_size]
        self.all_rating = tf.matmul(self.u_embedding, self.item_embeddings, transpose_a=False,
                                    transpose_b=True) + self.item_bias

        self.all_logits = tf.reduce_sum(tf.multiply(self.u_embedding, self.item_embeddings), 1) + self.item_bias
        self.NLL = -tf.reduce_mean(tf.log(
            tf.gather(tf.reshape(tf.nn.softmax(tf.reshape(self.all_logits, [1, -1])), [-1]), self.i))
        )
        # for dns sample
        self.dns_rating = tf.reduce_sum(tf.multiply(self.u_embedding, self.item_embeddings), 1) + self.item_bias


class IRGAN(AbstractRecommender):
    properties = [
        "factors_num",
        "lr",
        "g_reg",
        "d_reg",
        "epochs",
        "g_epoch",
        "d_epoch",
        "batch_size",
        "d_tau",
        "topk",
        "pretrain_file"
    ]

    def __init__(self, sess, dataset):
        # super(IRGAN, self).__init__()
        super().__init__(**kwds)

        train_matrix = dataset.trainMatrix.tocsr()
        self.num_users, self.num_items = train_matrix.shape

        self.factors_num = self.conf["factors_num"]
        self.lr = self.conf["lr"]
        self.g_reg = self.conf["g_reg"]
        self.d_reg = self.conf["d_reg"]
        self.epochs = self.conf["epochs"]
        self.g_epoch = self.conf["g_epoch"]
        self.d_epoch = self.conf["d_epoch"]
        self.batch_size = self.conf["batch_size"]
        self.d_tau = self.conf["d_tau"]
        self.topK = self.conf["topk"]
        self.pretrain_file = self.conf["pretrain_file"]
        self.loss_function = "None"

        idx_value_dict = {}
        for idx, value in enumerate(train_matrix):
            if any(value.indices):
                idx_value_dict[idx] = value.indices

        self.user_pos_train = idx_value_dict

        self.num_users, self.num_items = dataset.num_users, dataset.num_items


        self.all_items = np.arange(self.num_items)

    def build_graph(self):
        file = reader.lines(self.pretrain_file)
        pretrain_params = pickle.load(file, encoding="latin")
        self.generator = GEN(self.num_items, self.num_users, self.factors_num, self.g_reg, param=pretrain_params,
                             learning_rate=self.lr)
        self.discriminator = DIS(self.num_items, self.num_users, self.factors_num, self.d_reg, param=None,
                                 learning_rate=self.lr)

    def get_train_data(self):
        users_list, items_list, labels_list = [], [], []
        train_users = list(self.user_pos_train.keys())
        with ThreadPoolExecutor() as executor:
            data = executor.map(self.get_train_data_one_user, train_users)
        data = list(data)
        for users, items, labels in data:
            users_list.extend(users)
            items_list.extend(items)
            labels_list.extend(labels)

        return users_list, items_list, labels_list

    def get_train_data_one_user(self, user):
        user_list, items_list, label_list = [], [], []
        pos = self.user_pos_train[user]

        rating = self.sess.run(self.generator.all_rating, {self.generator.u: [user]})
        rating = np.reshape(rating, [-1])
        rating = np.array(rating) / self.d_tau  # Temperature
        exp_rating = np.exp(rating)
        prob = exp_rating / np.sum(exp_rating)
        neg = np.random.choice(self.all_items, size=len(pos), p=prob)
        for i, j in zip(pos, neg):
            user_list.append(user)
            items_list.append(i)
            label_list.append(1.0)

            user_list.append(user)
            items_list.append(j)
            label_list.append(0.0)
        return (user_list, items_list, label_list)

    def train_model(self):
        for _ in range(self.epochs):
            for _ in range(self.d_epoch):
                users_list, items_list, labels_list = self.get_train_data()
                self.training_discriminator(users_list, items_list, labels_list)
            for _ in range(self.g_epoch):
                self.training_generator()
                Evaluate.test_model(self, self.dataset)

    def training_discriminator(self, user, item, label):
        num_training_instances = len(user)
        for num_batch in np.arange(int(num_training_instances / self.batch_size)):
            bat_users, bat_items, bat_lables = \
                data_gen._get_pointwise_batch_data(user, item, label, num_batch, self.batch_size)

            feed = {self.discriminator.u: bat_users,
                    self.discriminator.i: bat_items,
                    self.discriminator.label: bat_lables}
            self.sess.run(self.discriminator.d_updates, feed_dict=feed)

    def training_generator(self):
        for user, pos in self.user_pos_train.items():
            sample_lambda = 0.2
            rating = self.sess.run(self.generator.all_logits, {self.generator.u: user})
            exp_rating = np.exp(rating)
            prob = exp_rating / np.sum(exp_rating)  # prob is generator distribution p_\theta

            pn = (1 - sample_lambda) * prob
            pn[pos] += sample_lambda * 1.0 / len(pos)
            # Now, pn is the Pn in importance sampling, prob is generator distribution p_\theta

            sample = np.random.choice(self.all_items, 2 * len(pos), p=pn)
            ###########################################################################
            # Get reward and adapt it with importance sampling
            ###########################################################################
            feed = {self.discriminator.u: user, self.discriminator.i: sample}
            reward = self.sess.run(self.discriminator.reward, feed_dict=feed)
            reward = reward * prob[sample] / pn[sample]
            ###########################################################################
            # Update G
            ###########################################################################
            feed = {self.generator.u: user, self.generator.i: sample, self.generator.reward: reward}
            self.sess.run(self.generator.gan_updates, feed_dict=feed)

    def predict(self, user_id, items):
        user_embedding, item_embedding, item_bias = self.sess.run(self.generator.g_params)
        u_embedding = user_embedding[user_id]
        item_embedding = item_embedding[items]
        item_bias = item_bias[items]

        ratings = np.matmul(u_embedding, item_embedding.T) + item_bias
        return ratings



'''
Xiangnan He et al., "Outer Product-based Neural Collaborative Filtering", In IJCAI 2018.
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.evaluation import Evaluate
from neurec.util import data_gen, learner
from neurec.util.properties import Properties

class ConvNCF(AbstractRecommender):
    properties = [
        "embedding_size",
        "topk",
        "regs",
        "keep",
        "epochs",
        "batch_size",
        "net_channel",
        "lr_embed",
        "lr_net",
        "verbose",
        "loss_function"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.embedding_size = self.conf["embedding_size"]
        self.topK = self.conf["topk"]
        self.regs = self.conf["regs"]
        self.lambda_bilinear =  self.regs[0]
        self.gamma_bilinear =  self.regs[1]
        self.lambda_weight =  self.regs[2]
        self.keep = self.conf["keep"]
        self.num_epochs= self.conf["epochs"]
        self.batch_size= self.conf["batch_size"]
        self.nc=self.conf["net_channel"]
        self.lr_embed=self.conf["lr_embed"]
        self.lr_net=self.conf["lr_net"]
        self.verbose= self.conf["verbose"]
        self.loss_function = self.conf["loss_function"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items
    
    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input_pos = tf.placeholder(tf.int32, shape = [None,], name = "item_input_pos")
            self.item_input_neg = tf.placeholder(tf.int32, shape = [None,], name = "item_input_neg")
            self.keep_prob = tf.placeholder(tf.float32, name = "keep_prob")
    #---------- model definition -------
    def weight_variable(self,shape):
        initial = tf.truncated_normal(shape, stddev=0.01)
        return tf.Variable(initial)

    def bias_variable(self,shape):
        initial = tf.constant(0.1, shape=shape)
        return tf.Variable(initial)

    def _conv_weight(self, isz, osz):
        return (self.weight_variable([2,2,isz,osz]), self.bias_variable([osz]))

    def _conv_layer(self, input, P):
        conv = tf.nn.conv2d(input, P[0], strides=[1, 2, 2, 1], padding='SAME')
        return tf.nn.tanh(conv + P[1])

    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.embedding_P = tf.Variable(tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.1),
                                                                name='embedding_P', dtype=tf.float32)  #(users, embedding_size)
            self.embedding_Q = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.1),
                                                                name='embedding_Q', dtype=tf.float32)  #(items, embedding_size)

            # here should have 6 iszs due to the size of outer products is 64x64
            iszs = [1] + self.nc[:-1]
            oszs = self.nc
            self.P = []
            for isz, osz in zip(iszs, oszs):
                self.P.append(self._conv_weight(isz, osz))

            self.W = self.weight_variable([self.nc[-1], 1])#32x1
            self.b = self.weight_variable([1])#1

    def _create_inference(self, item_input):
        with tf.name_scope("inference"):
            # embedding look up
            self.embedding_p = tf.nn.embedding_lookup(self.embedding_P, self.user_input)
            self.embedding_q = tf.nn.embedding_lookup(self.embedding_Q, item_input)

            # outer product of P_u and Q_i
            self.relation = tf.matmul(tf.expand_dims(self.embedding_p,2), tf.expand_dims(self.embedding_q,1))
            self.net_input = tf.expand_dims(self.relation, -1)

            # CNN
            self.layer = []
            input = self.net_input
            for p in self.P:
                self.layer.append(self._conv_layer(input, p))
                input = self.layer[-1]
            # prediction
            self.dropout = tf.nn.dropout(self.layer[-1], self.keep_prob)
            self.output_layer = tf.matmul(tf.reshape(self.dropout,[-1,self.nc[-1]]), self.W) + self.b

            return self.embedding_p, self.embedding_q, self.output_layer

    def _regular(self, params):
        res = 0
        for param in params:
            res += tf.reduce_sum(tf.square(param[0])) + tf.reduce_sum(tf.square(param[1]))
        return res

    def _create_loss(self):
        with tf.name_scope("loss"):
            # BPR loss for L(Theta)
            self.p1, self.q1, self.output = self._create_inference(self.item_input_pos)
            self.p2, self.q2, self.output_neg = self._create_inference(self.item_input_neg)
            self.result = self.output - self.output_neg
            self.loss = learner.pairwise_loss(self.loss_function, self.result)########

            self.opt_loss = self.loss + self.lambda_bilinear * ( tf.reduce_sum(tf.square(self.p1)) \
                                    + tf.reduce_sum(tf.square(self.q2)) + tf.reduce_sum(tf.square(self.q1)))\
                                    + self.gamma_bilinear * self._regular([(self.W, self.b)]) \
                                    + self.lambda_weight * (self._regular(self.P) + self._regular([(self.W, self.b)]))

    # used at the first time when emgeddings are pretrained yet network are randomly initialized
    # if not, the parameters may be NaN.
    def _create_pre_optimizer(self):
        self.pre_opt = tf.train.AdagradOptimizer(learning_rate=0.01).minimize(self.loss)

    def _create_optimizer(self):
        with tf.name_scope("learner"):
        # seperated optimizer
            var_list1 = [self.embedding_P, self.embedding_Q]
            #[self.W1,self.W2,self.W3,self.W4,self.b1,self.b2,self.b3,self.b4,self.P1,self.P2,self.P3]
            var_list2 = list(set(tf.trainable_variables()) - set(var_list1))
            opt1 = tf.train.AdagradOptimizer(self.lr_embed)
            opt2 = tf.train.AdagradOptimizer(self.lr_net)
            grads = tf.gradients(self.opt_loss, var_list1 + var_list2)
            grads1 = grads[:len(var_list1)]
            grads2 = grads[len(var_list1):]
            train_op1 = opt1.apply_gradients(zip(grads1, var_list1))
            train_op2 = opt2.apply_gradients(zip(grads2, var_list2))
            self.optimizer = tf.group(train_op1, train_op2)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_pre_optimizer()
        self._create_optimizer()
    #---------- training process -------
    def train_model(self):
        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input_pos, item_input_neg = data_gen._get_pairwise_all_data(self.dataset)
            total_loss = 0.0
            training_start_time = time()
            num_training_instances = len(user_input)
            for num_batch in np.arange(int(num_training_instances/self.batch_size)):
                bat_users,bat_items_pos,bat_items_neg =\
                 data_gen._get_pairwise_batch_data(user_input,\
                 item_input_pos, item_input_neg, num_batch, self.batch_size)
                feed_dict = {self.user_input:bat_users,self.item_input_pos:bat_items_pos,\
                            self.item_input_neg:bat_items_neg,self.keep_prob:0.8}
                #out_put, out_put_neg = self.sess.run((self.output, self.output_neg), feed_dict=feed_dict)
                #self.logger.info("out_put:\t", out_put)
                #self.logger.info("out_put_neg:\t", out_put_neg)

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss
            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset,epoch)

    def predict(self, user_id, items):
        users = np.full(len(items), user_id, dtype=np.int32)
        return self.sess.run(self.output, feed_dict={self.user_input: users, self.item_input_pos: items,self.keep_prob:1.0})



'''
Reference: ThaiBinh Nguyen, et al. "NPE: Neural Personalized Embedding for Collaborative Filtering" in ijcai2018
@author: wubin
'''
from __future__ import absolute_import
from __future__ import division
import os
from neurec.model.AbstractRecommender import AbstractRecommender
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'

import tensorflow as tf
import numpy as np
from time import time
from neurec.util import learner, data_gen
from neurec.evaluation import Evaluate
from neurec.util.properties import Properties

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '2'
class NPE(AbstractRecommender):
    properties = [
        "learning_rate",
        "embedding_size",
        "learner",
        "loss_function",
        "topk",
        "epochs",
        "reg",
        "batch_size",
        "high_order",
        "verbose",
        "num_neg"
    ]

    def __init__(self, **kwds):
        super().__init__(**kwds)

        self.learning_rate = self.conf["learning_rate"]
        self.embedding_size = self.conf["embedding_size"]
        self.learner = self.conf["learner"]
        self.loss_function = self.conf["loss_function"]
        self.topK = self.conf["topk"]
        self.num_epochs= self.conf["epochs"]
        self.reg = self.conf["reg"]
        self.batch_size= self.conf["batch_size"]
        self.high_order = self.conf["high_order"]
        self.verbose= self.conf["verbose"]
        self.num_negatives= self.conf["num_neg"]
        self.num_users = self.dataset.num_users
        self.num_items = self.dataset.num_items

    def _create_placeholders(self):
        with tf.name_scope("input_data"):
            self.user_input = tf.placeholder(tf.int32, shape = [None,], name = "user_input")
            self.item_input = tf.placeholder(tf.int32, shape = [None,], name = "item_input")
            self.item_input_recents = tf.placeholder(tf.int32, shape = [None,None], name = "item_input_recents")
            self.lables = tf.placeholder(tf.float32, shape=[None,],name="labels")
    def _create_variables(self):
        with tf.name_scope("embedding"):
            self.embeddings_UI = tf.Variable(tf.truncated_normal(shape=[self.num_users, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_UI', dtype=tf.float32)  #(users, embedding_size)
            self.embeddings_IU = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_IU', dtype=tf.float32)  #(items, embedding_size)
            self.embeddings_IL = tf.Variable(tf.truncated_normal(shape=[self.num_items, self.embedding_size], mean=0.0, stddev=0.01),
                name='embeddings_IL', dtype=tf.float32)
    def _create_inference(self):
        with tf.name_scope("inference"):
            # embedding look up
            embeddings_UI_u = tf.nn.embedding_lookup(self.embeddings_UI, self.user_input)
            embeddings_IU_i = tf.nn.embedding_lookup(self.embeddings_IU,self.item_input)
            embeddings_LI_l = tf.nn.embedding_lookup(self.embeddings_IL, self.item_input_recents)

            context_embedding = tf.reduce_sum(embeddings_LI_l,1)
            predict_vector = tf.multiply(tf.nn.relu(embeddings_UI_u), tf.nn.relu(embeddings_IU_i))\
               +tf.multiply(tf.nn.relu(embeddings_IU_i),tf.nn.relu(context_embedding))
            predict = tf.reduce_sum(predict_vector,1)
            return embeddings_UI_u,embeddings_IU_i,embeddings_LI_l, predict


    def _create_loss(self):
        with tf.name_scope("loss"):
            UI_u,IU_i,LI_l,self.output = self._create_inference()
            self.loss = learner.pointwise_loss(self.loss_function,self.lables,self.output) + self.reg* (tf.reduce_sum(tf.square(UI_u)) \
            +tf.reduce_sum(tf.square(IU_i))+tf.reduce_sum(tf.square(LI_l)))

    def _create_optimizer(self):
        with tf.name_scope("learner"):
            self.optimizer = learner.optimizer(self.learner, self.loss, self.learning_rate)

    def build_graph(self):
        self._create_placeholders()
        self._create_variables()
        self._create_loss()
        self._create_optimizer()
#---------- training process -------
    def train_model(self):
        for epoch in  range(self.num_epochs):
            # Generate training instances
            user_input, item_input,item_input_recents, lables = data_gen._get_pointwise_all_highorder_data(self.dataset,self.high_order, self.num_negatives)

            num_training_instances = len(user_input)
            total_loss = 0.0
            training_start_time = time()

            for num_batch in np.arange(int(num_training_instances/self.batch_size)):

                bat_users, bat_items, bat_items_recents, bat_lables =\
                data_gen._get_pointwise_batch_seqdata(user_input, \
                item_input,item_input_recents, lables, num_batch, self.batch_size)
                feed_dict = {self.user_input:bat_users, self.item_input:bat_items,
                             self.item_input_recents:bat_items_recents,self.lables:bat_lables}

                loss,_ = self.sess.run((self.loss,self.optimizer),feed_dict=feed_dict)
                total_loss+=loss

            self.logger.info("[iter %d : loss : %f, time: %f]" %(epoch+1,total_loss/num_training_instances,time()-training_start_time))
            if epoch %self.verbose == 0:
                Evaluate.test_model(self,self.dataset)
    def predict(self, user_id, items):
        cand_items = self.dataset.trainDict[user_id]
        item_recents = []
        for _ in range(len(items)):
            item_recents.append( cand_items[len(cand_items)-self.high_order:])
        users = np.full(len(items), user_id, dtype='int32')
        return self.sess.run((self.output), feed_dict={self.user_input: users,\
                                        self.item_input_recents:item_recents, self.item_input: items})



from abc import ABC, abstractmethod
from neurec.data.Dataset import Dataset
from neurec.util.properties import Properties
import logging

class AbstractRecommender(ABC):
    """Abstract class for building a Recommender class."""
    @property
    @abstractmethod
    def properties(self):
        pass

    def __init__(self, sess):
        """Sets up the model with properties, dataset, and session."""
        self.logger = logging.getLogger(__name__)
        self.conf = Properties().getProperties(self.properties)
        self.dataset = Dataset()
        self.sess = sess

        self.logger.info("Arguments: %s " %(self.conf))

    @abstractmethod
    def build_graph(self):
        pass

    @abstractmethod
    def train_model(self):
        pass

    @abstractmethod
    def predict(self):
        pass



from neurec.evaluation.leaveoneout import HR
from neurec.evaluation.leaveoneout import NDCG as looNDCG
from neurec.evaluation.leaveoneout import AUC as looAUC
import heapq  # for retrieval topK
import numpy as np
from concurrent.futures import ThreadPoolExecutor


def evaluate_by_loo(model,evaluateMatrix,evaluateNegatives,num_thread):
    """
    Evaluate the performance (Hit_Ratio, NDCG) of top-K recommendation
    Return: score of each test rating.
    """
    global _model
    global _trainMatrix
    global _evaluateMatrix
    global _evaluateNegatives
    global _K
    _model = model
    _trainMatrix = _model.dataset.trainMatrix.tocsr()
    _evaluateMatrix = evaluateMatrix.tocsr()
    _evaluateNegatives = evaluateNegatives
    _K = _model.topK
    hits, ndcgs,aucs = [], [],[]
    _evaluateusers = []
    for u in range(_model.num_users):
        items_test = _evaluateMatrix[u].indices
        if len(items_test) >0:
            _evaluateusers.append(u)
    if(num_thread > 1): # Multi-thread
        with ThreadPoolExecutor() as executor:
            res = executor.map(eval_by_loo_user, _evaluateusers)
        res = list(res)
        with ThreadPoolExecutor() as executor:
            res = executor.map(eval_by_loo_user, range(_evaluateMatrix.shape[0]))
        res = list(res)
        hits = [r[0] for r in res]
        ndcgs = [r[1] for r in res]
        aucs = [r[2] for r in res]
    # Single thread
    else:
        # Single thread
        for u in _evaluateusers:
            if len(_evaluateMatrix[u].indices) !=0:
                (hr, ndcg,auc) = eval_by_loo_user(u)
                aucs.append(auc)
                hits.append(hr)
                ndcgs.append(ndcg)

    return (hits, ndcgs,aucs)

def eval_by_loo_user(u):
    target_item = _evaluateMatrix[u].indices[0]
    eval_items =[]
    if _evaluateNegatives != None:
        eval_items = _evaluateNegatives[u]
    else :
        all_items = set(np.arange(_model.num_items))
        eval_items = list(all_items - set(_trainMatrix[u].indices))
    eval_items.append(target_item)
    # Get prediction scores
    map_item_score = {}
    predictions = _model.predict(u,eval_items)
    for i in np.arange(len(eval_items)):
        item = eval_items[i]
        map_item_score[item] = predictions[i]
    # Evaluate top rank list
    ranklist = heapq.nlargest(_K, map_item_score, key=map_item_score.get)
    hr = HR.getHitRatio(ranklist, target_item)
    ndcg = looNDCG.getNDCG(ranklist, target_item)
    auc = looAUC.getAUC(predictions, _K)
    return (hr,ndcg,auc)
